# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.

"""Statistic extraction and spatial disaggregation tools."""

import logging
import numpy as np

import pandas as pd
import rasterio

from inca.common.geoprocessing import block_window_generator, number_blocks
from inca.common.nuts import NUTS_ID
from inca.common.ecosystem import ECOTYPE

SUM = 'sum'
COUNT = 'px_count'


logger = logging.getLogger(__name__)


def statistics_byArea(path_data_raster, path_area_raster, area_names,
                      add_progress=lambda p: None, block_shape=(2048, 2048)):
    """Extract sum and count statistics for all areas given in the area_raster for the data_raster.

    Note: currently only works correctly for raster with absolute data values (relative datasets would need unit and pixel
    area)

    :param path_data_raster: path to the raster dataset containing the absolute values
    :param path_area_raster: path to the raster holding the areas for which statistics are extracted
    :param area_names: dict or Series mapping raster value to clear name of areas (key/index = clear name,
                       value = raster value)
    :param add_progress: callback function to update progress bar.
    :param block_shape: tuple (y_size, x_size) for block processing of the statistic extraction (optional)
    :return: pandas dataframe with area codes as index and columns for raster value sum and raster pixel count
    """

    # ini pandas DataFrame to hold the results
    df = pd.DataFrame(0., index=area_names.keys(), columns=[SUM, COUNT], dtype=float)
    df.index.name = NUTS_ID

    with rasterio.open(path_data_raster) as src_data, \
            rasterio.open(path_area_raster) as src_area:
        # extract raster info for progress bar
        src_profile = src_data.profile
        # number of to process blocks
        nblocks = number_blocks(src_profile, block_shape)

        # now iterate over the blocks of src files and process
        for _, window in block_window_generator(block_shape, src_profile['height'], src_profile['width']):
            # read data; convert to float to avoid overflow when original data is an integer type
            aData = src_data.read(1, window=window, masked=True).astype(float)
            aArea = src_area.read(1, window=window)

            # check if empty
            if aData.mask.all() or (aArea == 0).all():
                add_progress(100. / nblocks)
                continue
            # reduce workload by only looping over existing areas & ETclasses in the block
            lAreas = np.unique(aArea[(aArea != 0) & (~aData.mask)]).tolist()

            # extract sums and pixel count for each area
            for pArea, area_raster_value in area_names.items():
                if area_raster_value not in lAreas:
                    continue
                df.loc[pArea, SUM] += aData[aArea == area_raster_value].filled(0).sum()
                df.loc[pArea, COUNT] += ((aArea == area_raster_value) & (~aData.mask)).sum()
            add_progress(100. / nblocks)
    return df


def statistics_byArea_byET(path_data_raster, path_area_raster, area_names, path_ET_raster, ET_names,
                           add_progress=lambda p: None,
                           block_shape=(2048, 2048)):
    """Extract sum and count statistics for all Et given by ET_raster in areas given in the area_raster for the data_raster.

    Note: currently only works correctly for raster with absolute data values (relative datasets would need
    unit and pixel area)

    :param path_data_raster: path to the raster dataset containing the absolute values
    :param path_area_raster: path to the raster holding the areas for which statistics are extracted
    :param area_names: dict or Series mapping raster value to clear name of areas (key/index = clear name/area,
                       value = raster value)
    :param path_ET_raster: path to raster dataset containing the ET classes for sub-grouping
    :param ET_names: dict  or Series mapping raster value to ecosystem types (key/index = clear name/ecosystem type,
                     value = raster value)
    :param add_progress: callback function to update progress bar.
    :param block_shape: tuple (y_size, x_size) for block processing of the statistic extraction (optional)
    :return: pandas dataframe with area codes and ET class as multi-index and columns for raster value sum and
             raster pixel count
    """

    # ini pandas DataFrame to hold the results
    index = pd.MultiIndex.from_product([area_names.keys(), ET_names.keys()], names=[NUTS_ID, ECOTYPE])
    df = pd.DataFrame(0., index=index, columns=[SUM, COUNT], dtype=float)

    with rasterio.open(path_data_raster) as src_data, \
            rasterio.open(path_area_raster) as src_area, \
            rasterio.open(path_ET_raster) as src_ET:
        # extract raster info for progress bar
        src_profile = src_data.profile
        # number of to process blocks
        nblocks = number_blocks(src_profile, block_shape)

        # now iterate over the blocks of src files and process
        for _, window in block_window_generator(block_shape, src_profile['height'], src_profile['width']):
            # read data
            aData = src_data.read(1, window=window, masked=True).astype(float)
            aArea = src_area.read(1, window=window)
            aET = src_ET.read(1, window=window)

            # check if empty
            if aData.mask.all() or (aArea == 0).all():
                add_progress(100. / nblocks)
                continue
            # reduce workload by only looping over existing areas & ETclasses in the block
            lAreas = np.unique(aArea[(aArea != 0) & (~aData.mask)]).tolist()
            lET = np.unique(aET[(aArea != 0) & (~aData.mask)]).tolist()

            # extract sums and pixel count for each area/ETclass combination
            for pArea, area_raster_value in area_names.items():
                if area_raster_value not in lAreas:
                    continue
                for pET, eco_raster_value in ET_names.items():
                    if eco_raster_value not in lET:
                        continue
                    df.loc[(pArea, pET), SUM] += aData[(aArea == area_raster_value) &
                                                       (aET == eco_raster_value)].filled(0).sum()
                    df.loc[(pArea, pET), COUNT] += ((aArea == area_raster_value) &
                                                    (aET == eco_raster_value) & (~aData.mask)).sum()
            add_progress(100. / nblocks)
    return df


def spatial_disaggregation_byArea(path_proxy_raster, data, path_area_raster, area_names, path_out,
                                  add_progress=lambda p: None,
                                  proxy_sums=None,
                                  metadata=None, processing_info='N/A', unit_info='N/A',
                                  block_shape=(2048, 2048)):
    """Given an Area GeoTiff and area-specific total values, disaggregate the totals according to the proxy raster, such
    that the sum of the disaggregated values for the pixels within each Area equals the given total for the area. Uses
    block processing to reduce memory usage.

    :param path_proxy_raster: path to the raster dataset containing the proxy for spatially disaggregation
    :param data: Series containing the area-specific total values to distribute, indexed by area codes
    :param path_area_raster: path to the raster holding the areas for which statistics are extracted
    :param area_names: dict or Series mapping raster value to clear name of areas (key/index = clear name,
                       value = raster value)
    :param path_out: absolute output file name for spatially disaggregated raster file
    :param add_progress: callback function to update progress bar.
    :param proxy_sums: Optional Series of precomputed sum of proxy data per Area.
    :param metadata: METADATA object holding the extent and profile for reference GeoTiff file; if not given the info
                     is extracted from the proxy file
    :param processing_info: string describing the processing step (optional)
    :param unit_info:  string describing the unit of the spatially disaggregated totals (optional)
    :param block_shape: tuple (y_size, x_size) for block processing of the statistic extraction (optional)
    :return: DataFrame containing sum of proxy values per region.  Can be reused in subsequent calls to
             :meth:`spatial_disaggregation_byArea` to save computation effort.

    """
    progress_remain = 100.  # remaining progress to be used for progress bar
    if proxy_sums is None:
        # first get the proxy raster value sum of the different areas for which to disaggregate the table data
        proxy_sums = statistics_byArea(path_proxy_raster, path_area_raster, area_names,
                                       add_progress=lambda p: add_progress(0.4 * p),  # assign 40% of progress to stats
                                       block_shape=block_shape)[SUM]
        progress_remain = 60.  # 40% assigned to the statistics extraction

    # calculate area-specific contribution factor
    # TODO make sure area_names and data have the same index before calling statistics_byArea, so proxy_sums will
    #  be complete! (and the reindex is not needed here anymore)?  Take care that area_names can be a dict (convert to Series first?)!
    dis_factor = (data / proxy_sums.reindex(data.index)).fillna(0)

    # start processing
    with rasterio.open(path_proxy_raster) as src_proxy,\
            rasterio.open(path_area_raster) as src_area:

        # get info for block processing and output raster file
        if metadata is None:
            # get profile from proxy_file
            dst_profile = src_proxy.profile
        else:
            dst_profile = metadata.profile.copy()
        dst_profile.update(dtype=rasterio.float32, nodata=np.nan)
        # number of to process blocks
        nblocks = number_blocks(dst_profile, block_shape)

        # init the output raster file
        with rasterio.open(path_out, 'w', **dst_profile) as dst:
            # add metadata if needed
            if metadata is not None:
                metadata.read_raster_tags([path_proxy_raster, path_area_raster])
                dst.update_tags(**metadata.prepare_raster_tags(processing_info, unit_info))

            # now iterate over the blocks
            for _, window in block_window_generator(block_shape, dst_profile['height'], dst_profile['width']):
                # read data
                aProxy = src_proxy.read(1, window=window, masked=True)
                aArea = src_area.read(1, window=window)

                # init output block
                aData = np.full_like(aProxy, dst_profile['nodata'], dtype=dst_profile['dtype'])

                # check if empty
                if aProxy.mask.all() or (aArea == 0).all():
                    dst.write(aData, 1, window=window)
                    add_progress(progress_remain / nblocks)
                    continue
                # reduce workload by only looping over existing areas & ETclasses in the block
                lAreas = np.unique(aArea[(aArea != 0) & (~aProxy.mask)]).tolist()

                # loop over the areas to spatially disaggregate the data to the proxy
                for area_name, area_value in area_names.items():
                    if area_value not in lAreas:
                        continue
                    np.multiply(aProxy.data, dis_factor[area_name],
                                where=((aArea == area_value) & (~aProxy.mask)), out=aData)
                # write to disk
                dst.write(aData, 1, window=window)
                add_progress(progress_remain / nblocks)  # remaining progress here
    return proxy_sums  # Return proxy_sums so it may be reused by the caller in subsequent disaggregations


def spatial_disaggregation_byArea_byET(path_proxy_raster, data, path_area_raster, area_names,
                                       path_ET_raster, ET_names, path_out,
                                       add_progress=lambda p: None,
                                       proxy_sums=None,
                                       metadata=None, processing_info='N/A', unit_info='N/A', block_shape=(2048, 2048)):
    """Given an Area GeoTiff plus EcosystemType Geotiff and area/ET-specific total values, disaggregate the totals according
    to the proxy raster, such that the sum of the disaggregated values for the pixels for each ET class within each Area
    equals the given total for the ET class and area. Uses block processing to reduce memory usage.

    :param path_proxy_raster: path to the raster dataset containing the proxy for spatially disaggregation
    :param data: Series containing the area-ET-specific total values to distribute, MultiIndex is mandatory given first
                 the AreaID (NUTS_ID) and second the ETClassID (ECOTYPE).
    :param path_area_raster: path to the raster holding the areas for which statistics are extracted
    :param area_names: dict or Series mapping raster value to clear name of areas (key/index = clear name,
                       value = raster value)
    :param path_ET_raster: path to raster dataset containing the ET classes for sub-grouping
    :param ET_names: dict or Series mapping raster value to ecosystem types (key/index = ecosystem type,
                     value = raster value)
    :param path_out: absolute output file name for spatially disaggregated raster file
    :param add_progress: callback function to update progress bar.
    :param proxy_sums: Optional Series of precomputed sum of proxy data per Area and per ET.
    :param metadata: METADATA object holding the extent and profile for reference GeoTiff file; if not given the info
                     is extracted from the proxy file
    :param processing_info: string describing the processing step (optional)
    :param unit_info: string describing the unit of the spatially disaggregated totals (optional)
    :param block_shape: tuple (y_size, x_size) for block processing of the statistic extraction (optional)
    :return: DataFrame containing sum of proxy values per region and per ecosystem type.  Can be reused in subsequent
             calls to :meth:`spatial_disaggregation_byArea_byET` to save computation effort.

    """
    progress_remain = 100.
    if proxy_sums is None:
        # first get the proxy raster value sum of the different areas for which to disaggregate the table data
        proxy_sums = statistics_byArea_byET(path_proxy_raster, path_area_raster, area_names, path_ET_raster, ET_names,
                                            add_progress=lambda p: add_progress(0.4 * p),  # 40% of progress here
                                            block_shape=block_shape)[SUM]
        progress_remain = 60.

    # calculate area-specific contribution factor
    # Note: proxy_sums has a MultiIndex
    # TODO make sure all indices from data appear in area_names and ET_names before calling statistics_byArea_byET,
    #  so proxy_sums exist for every row in data (and the reindex is perhaps not needed here anymore)?
    #  Take care that area_names and ET_names can be dict (convert to Series first?)!
    dis_factor = (data / proxy_sums.reindex(data.index)).fillna(0)

    # start processing
    with rasterio.open(path_proxy_raster) as src_proxy, \
            rasterio.open(path_area_raster) as src_area, \
            rasterio.open(path_ET_raster) as src_ET:

        # get info for block processing and output raster file
        if metadata is None:
            # get profile from proxy_file
            dst_profile = src_proxy.profile
        else:
            dst_profile = metadata.profile.copy()
        dst_profile.update(dtype=rasterio.float32, nodata=np.nan)
        # number of to process blocks
        nblocks = number_blocks(dst_profile, block_shape)

        # init the output raster file
        with rasterio.open(path_out, 'w', **dst_profile) as dst:
            # add metadata if needed
            if metadata is not None:
                metadata.read_raster_tags([path_proxy_raster, path_area_raster, path_ET_raster])
                dst.update_tags(**metadata.prepare_raster_tags(processing_info, unit_info))

            # now iterate over the blocks
            for _, window in block_window_generator(block_shape, dst_profile['height'], dst_profile['width']):
                # read data
                aProxy = src_proxy.read(1, window=window, masked=True)
                aArea = src_area.read(1, window=window)
                aET = src_ET.read(1, window=window)

                # init output block
                aData = np.full_like(aProxy, dst_profile['nodata'], dtype=dst_profile['dtype'])

                # check if empty
                if aProxy.mask.all() or (aArea == 0).all():
                    dst.write(aData, 1, window=window)
                    add_progress(progress_remain / nblocks)
                    continue
                # reduce workload by only looping over existing areas & ETclasses in the block
                lAreas = np.unique(aArea[(aArea != 0) & (~aProxy.mask)]).tolist()
                lET = np.unique(aET[(aArea != 0) & (~aProxy.mask)]).tolist()

                # loop over the areas & ETclasses to spatially disaggregate the data to the proxy
                for area_name, area_value in area_names.items():
                    if area_value not in lAreas:
                        continue
                    for eco_type, eco_id in ET_names.items():
                        if eco_id not in lET:
                            continue
                        np.multiply(aProxy.data, dis_factor[area_name, eco_type],
                                    where=((aArea == area_value) & (aET == eco_id) & (~aProxy.mask)),
                                    out=aData)
                # write to disk
                dst.write(aData, 1, window=window)
                add_progress(progress_remain / nblocks)  # 60% of progress here
    return proxy_sums  # For optional reuse by the caller in subsequent disaggregations
