# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.
"""Crop provision."""

import sys

if sys.version_info[:2] >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources

import logging
import os

import numpy as np
import pandas as pd

import inca
import inca.common.estat as estat
import inca.common.suts as suts
from inca.common.config_check import ConfigItem, check_csv, check_tsv, ConfigRaster, YEARLY
from inca.common.errors import Error, ConfigError
from inca.common.ecosystem import ECOTYPE, ECOTYPE_L1, ecotype_l1_cat, types_l1, ECO_ID, SHAPE_ID, \
    CROP, GRASSLAND
from inca.common.geoprocessing import statistics_byArea_byET, RasterType
from inca.common.nuts import NUTS_ID

logger = logging.getLogger(__name__)

# some globals for faster adaptation
_CROPPROVISION = 'Crop Provision'
_OPERATION_MODE = 'operation-mode'
_LUT_CROPS = 'lut_crops'
_CROP_YIELD = 'crop_yield'
_PROXY_CROPTYPE = 'proxy_croptype'
_PROXY_CROPYIELD = 'proxy_cropyield'

# naming of calculation methods based on data source
MAT_FLOW = 'material-flow'
AGRI_STATS = 'agriculture-statistics'
GEOSPATIAL = 'agriculture-yield-map'

# TODO check
CODE_MFA = 'code_mfa'
AACT_UV = 'aact_uv01'
ECOCON = 'ecocon'

# units

EUR_TONNE = 'EUR_T'  # code for prices in euro per tonne in unit value table
EU_STD_HUMIDITY = 'PR_HU_EU'  # code for Harvested production in eu standard humidity (1000 t)
_UNIT = 'unit'
_THOUSAND_TONNE = 'THS_T'

_INDIC_ENV = 'indic_env'
_DOMESTIC_EXTRACTION = 'DE'


mfa_labels = {'MF11': 'MF1.1 crops (excluding fodder crops)',
              'MF111': 'MF1.1.1 cereals',
              'MF112': 'MF1.1.2 roots and tubers',
              'MF113': 'MF1.1.3 sugar crops',
              'MF114': 'MF1.1.4 pulses',
              'MF115': 'MF1.1.5 nuts',
              'MF116': 'MF1.1.6 oil-bearing crops',
              'MF117': 'MF1.1.7 vegetables',
              'MF118': 'MF1.1.8 fruits',
              'MF119': 'MF1.1.9 fibres',
              'MF11A': 'MF1.1.A other crops (excluding fodder crops) n.e.c.',
              'MF12':  'MF1.2 Crop residues (used) and fodder crops and grazed biomass',
              'MF121': 'MF1.2.1 Crop residues (used)',
              'MF1211': 'MF1.2.1.1 Straw',
              'MF1212': 'MF1.2.1.2 Other crop residues (sugar and fodder beet leaves and other)',
              'MF122': 'MF1.2.2 Fodder crops and grazed biomass',
              'MF1221': 'MF1.2.2.1 Fodder crops (including biomass harvest from grassland)',
              'MF1222': 'MF1.2.2.2 Grazed biomass'}

output_labels = {'SWHE': 'soft wheat',
                 'DWHE': 'durum wheat',
                 'BARL': 'barley',
                 'OATS': 'oats',
                 'MAIZ': 'maize',
                 'OCER': 'other cereals',
                 'RAPE': 'rape',
                 'SUNF': 'sunflower',
                 'MAIF': 'fodder maize',
                 'OFAR': 'other forage',
                 'PULS': 'protein crops',
                 'POTA': 'potatoes',
                 'SUGB': 'sugar beet'}


class CropRun(inca.Run):

    service_name = _CROPPROVISION

    def __init__(self, config):
        super().__init__(config)

        if config.get(_OPERATION_MODE) == MAT_FLOW:  # Currently no monetary evaluation for material flow
            self.monetary = False

        self.config_template.update({
            self.service_name: {
                _OPERATION_MODE: ConfigItem(default=MAT_FLOW),
                _LUT_CROPS: ConfigItem(check_csv, optional=True,
                                       required_columns=[CODE_MFA, 'label_mfa', ECOTYPE_L1]),
                _CROP_YIELD: ConfigItem(check_tsv),  # apro_cpsh1 table for AGRI_STATS, env_ac_mfa table for MAT_FLOW
                _PROXY_CROPTYPE: {YEARLY: ConfigRaster(optional=True, raster_type=RasterType.CATEGORICAL)},
                _PROXY_CROPYIELD: {YEARLY: ConfigRaster(optional=True, raster_type=RasterType.RELATIVE)}
            }
        })
        if self.monetary:  # Inputs used only for monetary evaluation:
            logger.debug('Add config_template for monetary')
            self.config_template[self.service_name].update({
                AACT_UV: ConfigItem(check_tsv),
            })

        # if a developer wants to run the original KIP-INCA JRC method, some parameters are to be overruled
        # this method is not available from QGIS and requires a command line option
        if self.config.get('jrcmethod'):
            logger.warning('jrcmethod overwrite set.  Operation mode is %s; crop_yield should be in apro_cpsh1 format.',
                           AGRI_STATS)
            self.config_template[self.service_name].update({
                    ECOCON: ConfigItem(check_csv, required_columns=[NUTS_ID] + list(output_labels))})

            self.config[self.service_name][_OPERATION_MODE] = AGRI_STATS
            self.config['statistics_shape'] = 'NUTS-0'
            self.config['reporting_shape'] = 'NUTS-0'
            self.config['selected_regions'] = ['EU27']

        self.lut_crops = None
        self.make_output_filenames()

    def _start(self):

        # preprocess statistics input table
        production, crops = self.preprocess_yield_statistics()

        # read unit values at basic prices:
        if self.monetary:
            self.unit_value = estat.read_tsv(self.config[self.service_name][AACT_UV])
            self.unit_value = self.unit_value.xs(EUR_TONNE, level='unit')
            # Manually replace 0. by pd.NA, as method .replace(0., pd.NA) throws RecursionError on some pandas versions.
            for col in self.unit_value.columns:
                self.unit_value.loc[self.unit_value[col] == 0., col] = pd.NA

        # filter on crops to account for operation mode: MFA, APRO, GEO # TODO
        if self.config.get('jrcmethod'):
            crops = self.lut_crops.dropna(subset=['code_CAPRI'])

        nyears = len(self.years)
        # loop over all processing years
        for year in self.years:
            logger.debug('year: %d', year)
            self.year = year

            # create empty frame for results
            self.result_tonne = pd.DataFrame(index=self.statistics_shape.index, columns=pd.Index([], name=CODE_MFA))
            self.result_eur = pd.DataFrame(index=self.statistics_shape.index, columns=pd.Index([], name=CODE_MFA))

            # extract production statistics per crop and map to MFA
            self.crop_statistics(production, crops)

            # map MFA production statistics to ecosystem types, prepare for reporting
            self.supply_tonne = self.reporting_to_ET()

            # Disaggregate statistics using provided proxy/proxies or default via ecosystem extent map
            # get links to proxy/proxies, if available
            proxy_raster_file_type = None
            proxy_raster_file_yield = None
            if self.config[_CROPPROVISION][_PROXY_CROPYIELD][self.year]:
                proxy_raster_file_yield = self.config[self.service_name][_PROXY_CROPYIELD][self.year]
            if self.config[_CROPPROVISION][_PROXY_CROPTYPE][self.year]:
                proxy_raster_file_type = self.config[self.service_name][_PROXY_CROPTYPE][self.year]
            # generate map based on statistics
            self.crop_provision_map_generation(proxy_raster_file_type, proxy_raster_file_yield,
                                               add_progress=lambda p: self.add_progress(0.5 * p / nyears))
            self.cut_to_reporting_regions(self.flow_map_physical[year])

            # Statistics extraction for reporting regions
            self.reporting_area_statistics(add_progress=lambda p: self.add_progress(0.5 * p / nyears))

            # Export of reporting area statistics to SUT Exel file and CSV statistic files
            self.reporting()

        return

    def make_output_filenames(self):
        """Create standard output file names."""
        crop_provision = _CROPPROVISION.lower().replace(' ', '-')  # lower case, without spaces
        self.flow_map_physical = {year: os.path.join(self.maps,
                                                     f'{crop_provision}_map_use_1000-tonne_{year}.tif')
                                  for year in self.years}
        self.flow_map_monetary_real = {year: os.path.join(self.maps,
                                                          f'{crop_provision}_map_use_M-EURO-real_{year}.tif')
                                       for year in self.years}
        self.flow_map_monetary_nominal = {year: os.path.join(self.maps,
                                                             f'{crop_provision}_map_use_M-EURO-current_{year}.tif')
                                          for year in self.years}
        self.use_physical = {year: os.path.join(self.statistic_dir, f'{crop_provision}_table_use_1000-tonne_{year}.csv')
                             for year in self.years}
        self.use_monetary = {year: os.path.join(self.statistic_dir,
                                                f'{crop_provision}_table_use_M-EURO-real_{year}.csv')
                             for year in self.years}
        self.use_monetary_nominal = {year: os.path.join(self.statistic_dir,
                                                        f'{crop_provision}_table_use_M-EURO-current_{year}.csv')
                                     for year in self.years}
        self.supply_physical = {year: os.path.join(self.statistic_dir,
                                                   f'{crop_provision}_table_supply_1000-tonne_{year}.csv')
                                for year in self.years}
        self.supply_monetary = {year: os.path.join(self.statistic_dir,
                                                   f'{crop_provision}_table_supply_M-EURO-real_{year}.csv')
                                for year in self.years}
        self.supply_monetary_nominal = {year: os.path.join(self.statistic_dir,
                                                           f'{crop_provision}_table_supply_M-EURO-current_{year}.csv')
                                        for year in self.years}
        self.sut_physical = {year: os.path.join(self.suts,
                                                f'{crop_provision}_report_SUT-physical_1000-tonne_{year}.xlsx')
                             for year in self.years}
        self.sut_monetary_real = {year: os.path.join(self.suts,
                                                     f'{crop_provision}_report_SUT-monetary_M-EURO-real_{year}.xlsx')
                                  for year in self.years}
        self.sut_monetary_nominal = {year: os.path.join(
            self.suts, f'{crop_provision}_report_SUT-monetary_M-EURO-current_{year}.xlsx')
            for year in self.years}

    def preprocess_yield_statistics(self):
        """Retrieve crop statistics and map to supply table format (MFA based)."""
        # read crop production:
        production = estat.read_tsv(self.config[self.service_name][_CROP_YIELD])

        # Check _CROP_YIELD has the right format:
        # For MAT_FLOW, crop yield table should have the env_ac_mfa.tsv format, for AGRI_STATS, we should have the
        # apro_cpsh1.tsv format.
        if self.config[_CROPPROVISION][_OPERATION_MODE] == MAT_FLOW:
            self.code = CODE_MFA
            expected_index_levels = ('indic_env', 'material', 'unit')
            standard_file = 'env_ac_mfa.tsv'
        elif self.config[_CROPPROVISION][_OPERATION_MODE] == AGRI_STATS:
            self.code = 'code_apro'
            expected_index_levels = ('crops', 'strucpro')
            standard_file = 'apro_cpsh1.tsv'
        elif self.config[_CROPPROVISION][_OPERATION_MODE] in [GEOSPATIAL]:
            raise ConfigError('operation-mode geospatial is not yet supported', [_CROPPROVISION, _OPERATION_MODE])
        else:
            raise ConfigError(f'operation-mode was set to "{self.config[_CROPPROVISION][_OPERATION_MODE]}", '
                              f'should be set to "{MAT_FLOW}" or "{AGRI_STATS}".',
                              [_CROPPROVISION, _OPERATION_MODE])
        # Check index has the expected levels
        if any(level not in production.index.names for level in expected_index_levels):
            raise ConfigError(f'Crop yield file "{self.config[self.service_name][_CROP_YIELD]}" does not have the '
                              f'correct format.  For operation mode {self.config[self.service_name][_OPERATION_MODE]}, '
                              f'the file should have the same structure as Eurostat standard dataset {standard_file}.',
                              [self.service_name, _CROP_YIELD])

        # read lookup crop mapping table from MFA to (optional) APRO and or GEO
        # If no file provided by user, use default LUT:
        ref = importlib_resources.files(inca.data).joinpath('cropprovision_lut.csv')
        file = self.config[self.service_name][_LUT_CROPS] if self.config[self.service_name][_LUT_CROPS] else ref
        with open(file, 'r') as f:
            lut_crops = pd.read_csv(f, comment='#', dtype={'code_val': str, ECOTYPE_L1: ecotype_l1_cat})

        if self.code not in lut_crops:
            raise ConfigError(f'Missing column in file "{self.config[self.service_name][_LUT_CROPS]}".  For method '
                              f'"{_OPERATION_MODE}" method, the crop lookup table must have a column "{self.code}".',
                              [_CROPPROVISION, _LUT_CROPS])
        self.lut_crops = lut_crops.set_index(self.code)

        # clean production statistics table
        assert (self.code in ('code_apro', CODE_MFA)), \
            'Crop provision service supports only material flows "env_ac_mfa" or agriculture statistics "apro_cpsh1"'
        if self.code == 'code_apro':
            production = production.xs(EU_STD_HUMIDITY, level='strucpro')
        elif self.code == CODE_MFA:
            # cross-sect dataframe Domestic Extraction, keep croptype and geo
            production = production.xs((_DOMESTIC_EXTRACTION, _THOUSAND_TONNE), level=(_INDIC_ENV, _UNIT))
            self.lut_crops[CODE_MFA] = self.lut_crops.index

        # subset based on reporting areas
        # TODO check if reindex also works for NUTS1-3
        production = production.reindex(self.statistics_shape.index, level=1, fill_value=np.nan)
        crops = production.index.unique(level=0)  # apro: level 0 == 'crops', mfa: level 0 == 'material'

        return production, crops

    def crop_statistics(self, production, crops):
        """Map crop statistics to supply table format (MFA based)."""
        # read ecological contribution (only in KIP-INCA JRC-method mode)
        if self.config.get('jrcmethod'):
            ecocon = pd.read_csv(self.config[self.service_name][ECOCON], comment='#').set_index(NUTS_ID)
            crops = crops.index
            logger.debug('ecocon:\n%s', ecocon)

        # extract statistics per crop
        for crop in crops:
            logger.debug('extract statistics for crop %s', crop)

            if (crop not in self.lut_crops.index) or (np.array(pd.isnull(self.lut_crops.loc[crop].code_mfa)).any()):
                logger.warning('Skipping crop %s: not in lookup table or not mapped to Material-flow', crop)
                continue

            if self.monetary and \
                    (not pd.isna(self.lut_crops.loc[crop].code_val)) and \
                    (self.lut_crops.loc[crop].code_val in self.unit_value.index):
                price = mean_price(self.unit_value.loc[self.lut_crops.loc[crop].code_val], self.year,
                                   self.lut_crops.loc[crop].label_val)  # [eur/tonne]
            else:
                # TODO need to find source for prices in euro-tonnes for material flow
                price = np.nan
            tmp = pd.DataFrame(index=estat.eu_members(2020))

            # KIP-INCA JRCmethod takes average over 3 years, and uses ecosystem contribution 'emergy' approach
            if self.config.get('jrcmethod'):
                # calculate mean over 3 years
                for y in (self.year - 1, self.year, self.year + 1):
                    if y in production.columns:
                        tmp[y] = production.loc[crop, y]
                    else:
                        logger.warning('No production value for crop %s in year %d',
                                       self.lut_crops.loc[crop].code_prod, y)
                prod = tmp.mean(axis=1)  # [1000 tonnes]
                crop_tonne = prod * ecocon[self.lut_crops.loc[crop].code_CAPRI]
            else:
                crop_tonne = production.loc[crop, self.year]  # in thousand tonnes

            # express monetary valuation in millions euro
            crop_eur = crop_tonne * price / 1000

            # append results to dataframes (one for physical and one for monetary)
            colname = self.lut_crops.loc[crop].code_mfa   # use code, transform later to label
            if not isinstance(colname, str):
                colname = colname.unique()[0]

            if colname in self.result_tonne.columns:  # if multiple crops go in one CAPRI category
                self.result_tonne[colname] = self.result_tonne[colname].add(crop_tonne, fill_value=0.)
                self.result_eur[colname] = self.result_eur[colname].add(crop_eur, fill_value=0.)
            else:
                self.result_tonne[colname] = crop_tonne
                self.result_eur[colname] = crop_eur

        # "real" value in terms of 2000 inflation index
        if self.monetary:
            self.result_eur_real = self.result_eur.multiply(self.deflator[2000] / self.deflator[self.year],
                                                            axis='index')
            logger.debug('deflator values:\n%s', self.deflator[2000] / self.deflator[self.year])

        # remove subtotal row, avoid double count
        mfa_key_subtotals = ['MF11', 'MF12', 'MF121', 'MF122']
        for mfa in mfa_key_subtotals:
            if mfa in self.result_tonne.columns:
                self.result_tonne.drop(mfa, axis=1, inplace=True)
            if self.monetary:
                if mfa in self.result_eur.columns:
                    self.result_eur.drop(mfa, axis=1, inplace=True)
                if mfa in self.result_eur_real.columns:
                    self.result_eur_real.drop(mfa, axis=1, inplace=True)

        # add missing MFA types
        mfa_keys = list(mfa_labels.keys())
        mfa_keys = [x for x in mfa_keys if x not in mfa_key_subtotals]
        for mfa in mfa_keys:
            if mfa not in self.result_tonne.columns:
                self.result_tonne[mfa] = 0
            if self.monetary:
                if mfa not in self.result_eur.columns:
                    self.result_eur[mfa] = 0
                if mfa not in self.result_eur_real.columns:
                    self.result_eur_real[mfa] = 0

        if self.monetary:
            logger.debug('result_eur:\n%s', self.result_eur.sum(axis=1))
        logger.debug('result_tonne:\n%s', self.result_tonne.sum(axis=1))

        return

    def crop_provision_map_generation(self, proxy_raster_file_type, proxy_raster_file_yield, add_progress):
        """Generate the crop provision maps for the statistical regions."""
        logger.info('start generation of crop provision map for year %s', self.year)

        # Create dataframe indexed by NUTS_ID and ecosystem type:
        df_data = self.supply_tonne.droplevel(CODE_MFA).set_index(ECOTYPE_L1, append=True)
        df_data = df_data.groupby([NUTS_ID, ECOTYPE_L1]).sum()
        # Make sure index contains all ecotypes, and rename to ECOTYPE instead of ECOTYPE_L1:
        idx = pd.MultiIndex.from_product([df_data.index.unique(NUTS_ID), types_l1[ECOTYPE]])
        df_data = df_data.reindex(idx, fill_value=0.)

        if not proxy_raster_file_yield:
            if proxy_raster_file_type is None and proxy_raster_file_yield is not None:
                logger.warning('Geospatial distribution through crop type map not yet supported, '
                               'use extent map only.')
            else:
                logger.warning('no proxy map provided for crop-types and/or crop-yield, '
                               'use default ecosystem extent map.')

            # TODO: Question: should the unit really in the map in 1000 tonnes?
            self.accord.spatial_disaggregation_byArea_byET(self.get_ecosystem_raster(self.year, level=1),
                                                           df_data['physical'],
                                                           self.statistics_raster,
                                                           self.statistics_shape[SHAPE_ID].to_dict(),
                                                           self.get_ecosystem_raster(self.year, level=1),
                                                           types_l1.set_index('name')[ECO_ID].to_dict(),
                                                           self.flow_map_physical[self.year],
                                                           add_progress=add_progress,
                                                           processing_info='Crop provision supply map for year '
                                                           f'{self.year} using the ecosystem extent map as proxy',
                                                           unit_info='1000 tonnes')
            # TODO monetary

        else:  # have proxy_raster_file_yield:
            if proxy_raster_file_type is not None:
                logger.warning('geospatial distribution through crop type map not yet supported, use yield map only')
            else:
                logger.warning('no proxy map provided for crop-types, use yield map only')

            # If we have monetary evaluation, ~60% of time is spent on that, and 40% on physical part.  Without monetary
            # evaluation, 100% of time is spent on the physical part.
            progress_weight_physical = 0.4 if self.monetary else 1.
            proxy_sums = self.accord.spatial_disaggregation_byArea_byET(
                proxy_raster_file_yield, df_data['physical'],
                self.statistics_raster, self.statistics_shape[SHAPE_ID],
                self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID],
                self.flow_map_physical[self.year],
                add_progress=lambda p: add_progress(progress_weight_physical * p),
                unit_info='1000 tonnes', processing_info='Spatial disaggreation of use.')
            if self.monetary:
                self.accord.spatial_disaggregation_byArea_byET(
                    proxy_raster_file_yield, df_data['monetary'],
                    self.statistics_raster, self.statistics_shape[SHAPE_ID],
                    self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID],
                    self.flow_map_monetary_nominal[self.year],
                    add_progress=lambda p: self.add_progress(0.3 * p),
                    proxy_sums=proxy_sums,
                    unit_info='Million EUR', processing_info='Spatial disaggregation of use.')
                self.accord.spatial_disaggregation_byArea_byET(
                    proxy_raster_file_yield, df_data['monetary_real'],
                    self.statistics_raster, self.statistics_shape[SHAPE_ID],
                    self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID],
                    self.flow_map_monetary_real[self.year],
                    add_progress=lambda p: self.add_progress(0.3 * p),
                    proxy_sums=proxy_sums,
                    unit_info='Million EUR (Reference year 2000 value)',
                    processing_info='Spatial disaggregation of use.')
        return

    def reporting(self):
        """Write statistics csv files and SUT spreadsheets."""
        logger.info('statistics export')
        logger.info('* CSV statistic files')

        # in CSV files we use the shortnames for ecosystem types
        ecotype_short_name = types_l1.set_index(ECOTYPE)['short_name']
        # merge with short names, and keep only physical and short_name columns
        result_tonne_csv = self.supply_tonne.merge(ecotype_short_name,
                                                   left_on=ECOTYPE_L1, right_index=True).sort_index()[['physical',
                                                                                                       'short_name']]

        # replace MFA codes by longer MFA labels
        result_tonne_csv.rename(index=mfa_labels, inplace=True)
        result_tonne_csv.index.rename('MFA', level=CODE_MFA, inplace=True)

        # replace column labels
        result_tonne_csv.columns = ['supply', 'ecosystem type']

        result_tonne_csv.to_csv(self.supply_physical[self.year])
        # self.reporting_eur.to_csv(self.use_monetary[self.year], index_label='NUTS_ID')
        # self.reporting_eur_real.to_csv(self.supply_monetary[self.year], index_label='NUTS_ID')

        logger.info('* SUT Excel files')
        # supply/use tables: all use is intermediate consumption by industries, all supply is from cropland
        suts.write_sut(self.sut_physical[self.year],
                       self.reporting_tonne.unstack(),
                       self.reporting_tonne.unstack().sum(axis=1).rename(suts.IND),
                       self.service_name, self.year, '1000 tonnes')
        if self.monetary:
            suts.write_sut(self.sut_monetary_nominal[self.year],
                           self.reporting_eur.unstack(),
                           self.reporting_eur.unstack().sum(axis=1).rename(suts.IND),
                           self.service_name, self.year, 'Million EUR (2000 value)')
            suts.write_sut(self.sut_monetary_real[self.year],
                           self.reporting_eur_real.unstack(),
                           self.reporting_eur_real.unstack().sum(axis=1).rename(suts.IND),
                           self.service_name, self.year, 'Million EUR (nominal)')

    def reporting_to_ET(self):
        """Map table crops to ecosystem types."""
        # Move CODE_MFA to index:
        df = self.result_tonne.stack().sort_index().rename('physical').to_frame()
        if self.monetary:
            df['monetary'] = self.result_eur.stack()
            df['monetary_real'] = self.result_eur_real.stack()

        # Create a mapping from CODE_MFA to ecosystem type:
        mfa_ecotype_mapping = self.lut_crops[[ECOTYPE_L1, CODE_MFA]].dropna().drop_duplicates(subset=CODE_MFA).dropna()
        mfa_ecotype_mapping.set_index(CODE_MFA, inplace=True)

        types_l1_idx = types_l1.set_index('name')
        cropland_id = types_l1_idx.loc[CROP, ECOTYPE]
        grassland_id = types_l1_idx.loc[GRASSLAND, ECOTYPE]

        # assign missing MFA to ET (MFA11x to cropland, MFA12x to grassland)
        for mfa in self.result_tonne.columns:
            if mfa not in mfa_ecotype_mapping.index:
                if mfa.startswith('MF11'):
                    mfa_ecotype_mapping.loc[mfa] = cropland_id
                elif mfa.startswith('MF12'):
                    mfa_ecotype_mapping.loc[mfa] = grassland_id
                else:
                    raise Error(f"Don't know which ecosystem type to assign to crop with MFA code {mfa}.")

        # add ecosystem type column to dataframe by joining with mfa_mapping
        return df.join(mfa_ecotype_mapping)

    def reporting_area_statistics(self, add_progress):
        """Extract statistics for the reporting regions."""
        logger.info('extract statistics for reporting regions')

        # If we have monetary evaluation, ~60% of time is spent on that, and 40% on physical part.  Without monetary
        # evaluation, 100% of time is spent on the physical part.
        progress_weight_physical = 0.4 if self.monetary else 1.

        ecosystem_raster = self.get_ecosystem_raster(self.year, level=1,
                                                     add_progress=lambda p: add_progress(
                                                         progress_weight_physical * 0.5 * p))

        self.reporting_tonne = statistics_byArea_byET(self.flow_map_physical[self.year],
                                                      self.reporting_raster, self.reporting_shape[SHAPE_ID],
                                                      ecosystem_raster,
                                                      types_l1.set_index('name')[ECO_ID],
                                                      lambda p: add_progress(progress_weight_physical * 0.5 * p))['sum']

        if self.monetary:
            self.reporting_eur = statistics_byArea_byET(self.flow_map_monetary_nominal[self.year],
                                                        self.reporting_raster, self.reporting_shape[SHAPE_ID],
                                                        ecosystem_raster,
                                                        types_l1.set_index('name')[ECO_ID],
                                                        lambda p: add_progress(0.3 * p))['sum']
            self.reporting_eur_real = statistics_byArea_byET(self.flow_map_monetary_real[self.year],
                                                             self.reporting_raster, self.reporting_shape[SHAPE_ID],
                                                             ecosystem_raster,
                                                             types_l1.set_index('name')[ECO_ID],
                                                             lambda p: add_progress(0.3 * p))['sum']


def set_cmdline_args(parser):
    """Implement additional command line arguments.

    :param parser: subparser object for CropProvision case in the __main__
    :return:
    """
    parser.add_argument('--jrcmethod', action='store_true',
                        help='To run the crop provision in original KIP-INCA JRC method (emergy function)')


def mean_price(unit_value, year, cropname):
    """Calculate the average of the price for (year-1, year, year+1) from the unit_value table for a crop.

    Gap-filling:
    - if no prices available for those years, use the price from the nearest year for the same member state.
    - if no price for that member state at all, use the EU average for that year.
    """
    assert (not (unit_value == 0.).any(axis=None))  # prices of 0. must be replaced by NA values here.
    tmp = pd.DataFrame(index=unit_value.index)
    for y in (year - 1, year, year + 1):
        if y in unit_value.columns:
            tmp[y] = unit_value[y]
        else:
            logger.warning('No data for year %d and crop %s in unit_value table.', y, cropname)
    result = tmp.mean(axis=1)
    eu_average = estat.filter_eu27(result).mean()
    # countries where 3-year average is 0 or missing:
    for member_state in result.index[result.isna()]:
        # get all available non-zero prices for that member state:
        ms_prices = unit_value.loc[member_state].dropna()
        if ms_prices.empty:  # no actual prices for this member state
            result.loc[member_state] = eu_average
            logger.debug('Use eu average price for %d %s %s: %f', year, member_state, cropname, eu_average)
        else:  # find the price from the nearest year
            idx_nearest_year = np.argmin(np.abs(ms_prices.index - year))
            result.loc[member_state] = ms_prices.iloc[idx_nearest_year]
            logger.debug('Use price from nearest year (%d) for %d %s %s: %f',
                         ms_prices.index[idx_nearest_year],
                         year, member_state, cropname, result.loc[member_state])

    return result
