# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""Preprocess the input land cover classification to incorporate road maps.

Note: the input raster file is only enhanced and in no way altered in resolution, extent or EPSG.
"""

import logging
import argparse
import os
import sys
import time

import rasterio

import inca
import inca.common.geoprocessing as geo

logger = logging.getLogger(__name__)

description = 'INCA Flood control land cover + roads preprocessing'

def main():
    start_time = time.time()
    options = parse_args()

    inca.set_up_console_logging(options.verbose)

    logger.info('START OF MODULE: %s', description)
    process_data(options, block_shape=(1024, 1024))
    logger.info('All processing successfully done. Have a nice day.')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: %s', description)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('path_in', help='Path to input land cover classification', metavar='landcover_file')
    parser.add_argument('year', help='Year of the land cover classification.')
    parser.add_argument('path_main_roads', help='Path to main road  raster.  '
                        'Secondary road raster must be in the same directory', metavar='roads')
    parser.add_argument('path_out_base', help='Path to output folder.', metavar='output_dir')
    parser.add_argument('-z', '--overwrite', action='store_true', help='overwrite existing output files.')
    parser.add_argument('--verbose', action='store_true', help='verbose output')

    return parser.parse_args()


def process_data(options, block_shape=(2048, 2048)):
    """Block process the original LC map and enhance with road maps."""
    os.makedirs(options.path_out_base, exist_ok=True)

    # create output file path
    path_out = os.path.join(os.path.normpath(options.path_out_base), f'landcover_with_roads_{options.year}.tif')

    if os.path.exists(path_out) and (not options.overwrite):
        logger.critical('Output file already exists. Please set the -z flag to overwrite.')
        sys.exit(2)

    try:
        logger.info('* start the block processing...')
        with rasterio.open(options.path_in) as src, \
             rasterio.open(options.path_main_roads) as src1, \
             rasterio.open(options.path_main_roads.replace("main",
                                                           "secondary")) as src2, \
             rasterio.open(path_out, 'w', **dict(src.profile,
                                                 tiled=True,
                                                 compress='LZW',
                                                 blockysize=256,
                                                 blockxsize=256,
                                                 count=1)) as dst:
            # set tags
            metadata = geo.Metadata("INCA Tool", "Floodcontrol pre-processing")
            metadata.update_dataset_tags(dst,
                                         f'Year {options.year} landcover + OSM roads map for flood control',
                                         'dimensionless', src.name, src1.name, src2.name)

            # now iterate over the blocks of src files and process
            for index, window in geo.block_window_generator(block_shape, dst.profile['height'], dst.profile['width']):
                # read data
                aOriginal = src.read(1, window=window)
                main_roads = src1.read(1, window=window)
                secondary_roads = src2.read(1, window=window)
                aOriginal[(main_roads == 1) | (secondary_roads == 1)] = 1
                dst.write(aOriginal, 1, window=window)
        logger.debug('done.')
    except Exception as e:
        # here we delete file to rerun after error
        if os.path.exists(path_out):
            os.remove(path_out)
        raise RuntimeError(f'generation of landcover maps with roads information for year {options.year} failed: {e}')


if __name__ == '__main__':
    # hack to get the proper logger when running the script directly with python
    logger = logging.getLogger('inca.floodcontrol.preprocessing')
    main()
