#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    Bring2AOI.py -i <file_path> -o <folder_path> -r <reference_file> [-m, <mode>] [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing raster datasets needed in the soil retention model.
    Raster data is resampled plus warped to the needed AOI and resolution (standard nearest neighbor).
    Note: original file naming is kept, only suffix is added.

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.soilretention.preprocessing.Bring2AOI -i <file_path> -o <folder_path> -r <ref_file>
    OR
    python3 Bring2AOI.py -i <file_path> -o <folder_path> -r <ref_file>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.

VERSION
    1.1 (2021-11-18)
"""

import logging
import optparse
import os
import time
import subprocess

import rasterio

import inca
import inca.common.geoprocessing as geo
import inca.soilretention.preprocessing as preprocessing

logger = logging.getLogger(preprocessing.__name__)


def main():
    start_time = time.time()
    # parse the arguments
    options = parse_args()

    # ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA soil retention file warp preprocessing')

    # init the metadata and raster extent of project
    metadata = geo.Metadata("Dr. Marcel Buchhorn for INCA project",
                            "Soil Retention pre-processing",
                            os.path.normpath(options.path_ref))
    os.makedirs(os.path.normpath(options.path_out), exist_ok=True)

    # run the warp to AOI
    logger.info('bring file to AOI')
    try:
        bring_2_AOI(options, metadata)
    except Exception:
        logger.exception('Error during processing to bring dataset to final AOI and resolution')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA soil retention file warp preprocessing')


def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.1")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')

    parser.add_option('-r', '--path_reference', help='Path to raster reference file for resampling. Needed.',
                      dest='path_ref')
    parser.add_option('-i', '--path_in', help='Path to input raster file. Needed.', dest='path_in')
    parser.add_option('-o', '--path_out', help='Path to output folder. Needed.', dest='path_out')
    parser.add_option('-m', '--mode', help='Resampling mode (standard: near).', dest='res_mode')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_ref) or (not os.path.isfile(os.path.normpath(options.path_ref))):
        parser.error("the -r argument for the reference raster is missing or the given path doesn't exist!!!")
    if (not options.path_in) or (not os.path.isfile(os.path.normpath(options.path_in))):
        parser.error("the -i argument for raster input data is missing or the given path doesn't exist!!!")
    if not options.res_mode:
        options.res_mode = 'near'

    if len(args) != 0:
        parser.error('too many arguments')

    return options


def bring_2_AOI(options, metadata):
    ''' function of warp original dataset to correct EPSG, resolution and AOI

    Note: I'm not using the function from the geoprocessing.py due to needed adaptations
    ToDo: improve the geoprocesing.py functions for warping to run also stand-alone'''

    # final path_out
    path_out = os.path.join(os.path.normpath(options.path_out),
                            os.path.basename(os.path.normpath(options.path_in)).rsplit('.')[0] +
                            '_INCA_{}m_EPSG{}.tif'.format(int(metadata.profile['transform'].a),
                                                          metadata.profile['crs'].to_epsg()))

    cmd = 'gdalwarp --config GDAL_CACHEMAX 256 -t_srs "{}" -te {} {} {} {} -tr {} {} -r {} -et 0 -nomd '\
          '-wo SAMPLE_STEPS=50 -wo SOURCE_EXTRA=5 -wo SAMPLE-GRID=YES -co COMPRESS=LZW -co TILED=YES '\
          '-co INTERLEAVE=BAND -co BIGTIFF=YES -multi -overwrite "{}" "{}"'.format(
              metadata.profile['crs'].to_string().replace('"', '\\"'),
              metadata.pextent.left,
              metadata.pextent.bottom,
              metadata.pextent.right,
              metadata.pextent.top,
              metadata.profile['transform'].a,
              abs(metadata.profile['transform'].e),
              options.res_mode,
              os.path.normpath(options.path_in),
              path_out)

    try:
        subprocess.check_call(cmd, shell=True)
    except Exception:
        raise OSError('Could not warp the raster file to the AOI')

    # infuse extra metadata
    metadata.read_raster_tags([os.path.normpath(options.path_in)])
    tags = metadata.prepare_raster_tags(
        'Original file: {}. Warped to EPSG, resolution and extent of INCA AOI'.format(
            os.path.basename(os.path.normpath(options.path_in))), '')

    with rasterio.open(path_out, 'r+') as dst:
        dst.update_tags(**tags)


if __name__ == '__main__':
    main()
