#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    maes.py -i <path_to_raster_file> -c <path_to_csv> -o <path_to_output_folder> -y <year> -r <reference_file>
    [-n, --nodata] [-z, --overwrite] [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing the input land cover classification to match the classes needed for the
    soil retention preprocessing module c-factor.

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.soilretention.preprocessing.maes -i <path_to_raster_file> -c <path_to_csv> -o <path_to_output_folder> 
    -y <year> -r <reference_file>

    OR

    python3 maes.py -i <path_to_raster_file> -c <path_to_csv> -o <path_to_output_folder> -y <year> -r <reference_file>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

VERSION
    1.2 (2023-03-02)
"""
################## Importing modules ##################
# standard
import logging
import optparse
import os
import math
import sys
import time

# extra official modules
import rasterio

# own modules
import inca
import inca.common.classification as classification
import inca.common.geoprocessing as geo
import inca.soilretention.preprocessing as preprocessing
############## END importing modules ##################

logger = logging.getLogger(preprocessing.__name__)

############## MAIN ###################################


def main():
    start_time = time.time()
    inca.set_up_console_logging(verbose=True)

    # parse the arguments
    options = parse_args()

    # change umask driver
    try:
        old_umask = os.umask(0o006)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA soil retention MAES preprocessing')

    # init the metadata
    accord = geo.GeoProcessing("INCA-tool",
                               "Soil Retention pre-processing",
                               os.path.normpath(options.path_out_base),
                               ProjectExtentTiff=os.path.normpath(options.path_ref),
                               GDAL_verbose=options.verbose)
    os.makedirs(os.path.normpath(options.path_out_base), exist_ok=True)

    # make sure we have the landcover map in correct specs
    logger.info('check input landcover map')
    try:
        path_classification = accord.AutomaticBring2AOI(options.path_in, secure_run=True)
    except Exception:
        logger.exception('Error during processing to bring dataset to final AOI and resolution')
        raise

    # process the input files in chunks to save memory
    try:
        # alter the block_shape to play with the memory requirements
        maes_processing(options, path_classification, accord, block_shape=(1024, 1024))
    except Exception:
        logger.exception('Error during block processing of the input files.')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA soil retention MAES preprocessing')


def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.2")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')
    parser.add_option('-i', '--in', help='Path to input raster classification file. Needed.', dest='path_in')
    parser.add_option('-c', '--csv', help='Path to reclassification list. Needed.', dest='path_csv')
    parser.add_option('-o', '--out', help='Path to output folder. Needed.', dest='path_out_base')
    parser.add_option('-y', '--year', help='Year of the land cover map. Needed.', dest='year')
    parser.add_option('-r', '--path_reference', help='Path to raster reference file for resampling. Needed.',
                      dest='path_ref')
    parser.add_option('-n', '--nodata', help='New nodata value.', dest='new_nodata')
    parser.add_option('-z', '--overwrite', action='store_true', default=False, help='overwrite existing output files.')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_in) or (not os.path.isfile(os.path.normpath(options.path_in))):
        parser.error("the -i argument for the input raster is missing or the given path doesn't exist!!!")
    if (not options.path_csv) or (not os.path.isfile(os.path.normpath(options.path_csv))):
        parser.error("the -c argument for the conversion list is missing or the given path doesn't exist!!!")
    if (not options.path_ref) or (not os.path.isfile(os.path.normpath(options.path_ref))):
        parser.error("the -r argument for the reference raster is missing or the given path doesn't exist!!!")
    if (not options.year):
        parser.error("the -y argument for the year of the to convert land cover map is missing or the given path doesn't exist!!!")
    if (not options.new_nodata):
        print('Warning. No new nodata value was given. The nodata value of the input raster will be used.')
        options.new_nodata = 'tbd'

    if len(args) != 0:
        parser.error('too many arguments')

    return options


def maes_processing(options, path_lc, accord, block_shape=(2048, 2048)):
    ''' function to block process the original LC map and transfer to a meas raster'''

    # get the dict for the reclassification
    logger.info('* reading the reclassification CSv file')
    try:
        dClasses = classification.CSV_2_dict(os.path.normpath(options.path_csv))
    except Exception:
        logger.exception('The needed reclassification dictionary could not be generated from the given CSV file')
        raise

    # create output file path
    path_out = os.path.join(os.path.normpath(options.path_out_base), 'Maes_soil-retention_{}.tif'.format(options.year))

    logger.info('* start the block processing...')
    try:
        # setup dst profile and metadata
        profile_dst = accord._load_profile(path_lc)['profile']
        # check if we have a new nodata value
        if options.new_nodata == 'tbd':
            # we take the input nodata value
            options.new_nodata = profile_dst['nodata']

        profile_dst.update(tiled=True,
                           compress='DEFLATE',
                           blockysize=256,
                           blockxsize=256,
                           count=1,
                           nodata=options.new_nodata)
        # create tags
        accord.metadata.read_raster_tags([os.path.normpath(options.path_in)])
        tags = accord.metadata.prepare_raster_tags(
            'MAES map for soil retention model for year {}. Generated with reclassification file: {}'.format(
                options.year, os.path.basename(os.path.normpath(options.path_csv))), 'categories')

        # set loop breaker if file already exists and overwrite flag is not set
        if (os.path.exists(path_out)) and (options.overwrite == False):
            logger.critical(
                'Output file is already existing. Please set the -z flag for overwriting. Processing will be ended.')
            sys.exit(2)

        # counter for verbose view  --> non-esential code block
        if options.verbose:
            max_counter = math.ceil(profile_dst['height'] * 1.0 / block_shape[0]) * math.ceil(
                profile_dst['width'] * 1.0 / block_shape[1])
            counter = 1
            pstatus = 0

        # run block processing
        with rasterio.open(os.path.normpath(path_lc)) as src:
            # create new file
            with rasterio.open(path_out, 'w', **profile_dst) as dst:
                # set tags
                dst.update_tags(**tags)

                # now iterate over the blocks of src files and process
                for index, window in inca.common.geoprocessing.block_window_generator(block_shape,
                                                                                      profile_dst['height'],
                                                                                      profile_dst['width']):
                    # read data
                    aOriginal = src.read(1, window=window)

                    # reclassify
                    aReclassified, dClasses = classification.reclassification(aOriginal, dClasses, src.nodata,
                                                                              options.new_nodata)

                    # write block out
                    dst.write(aReclassified, 1, window=window)

                    # free data
                    aOriginal = None
                    aReclassified = None

                    # update print statement in verbose view --> non-esential code block
                    if options.verbose:
                        progress = (counter / max_counter) * 100
                        if (progress - pstatus) >= 5:
                            pstatus = int(progress)
                            logger.debug('{} %'.format(pstatus))
                        counter += 1
        logger.debug('done.')
    except Exception:
        # here we delete file to rerun after error
        if os.path.exists(path_out):
            os.remove(path_out)
        raise IOError('generation of MAES map for year {} was not successful'.format(options.year))


if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger('inca.soilretention.preprocessing.maes')
    main()
