# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.

import argparse
import logging
import os
import sys
import time
import richdem as rd

import rasterio

import inca
import inca.common.geoprocessing as geo
import inca.woodprovision.preprocessing

logger = logging.getLogger(__name__)

description = 'INCA Tool wood provision DEM pre-processing.'


def main():
    start_time = time.time()

    options = parse_args()

    inca.set_up_console_logging(options.verbose)

    logger.info('START OF MODULE: %s', description)

    try:
        dem_processing(options, block_shape=(1024, 1024))
    except Exception:
        logger.exception('Error during block processing of the input files.')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA wood provision dem preprocessing')


def parse_args():
    parser = argparse.ArgumentParser(description)
    parser.add_argument('path_in', help='Path to input DEM raster file.', metavar='input_dem')
    parser.add_argument('path_out_base', help='Path to output folder', metavar='output_dir')
    parser.add_argument('-z', '--overwrite', action='store_true',
                        help='overwrite existing output files.')
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose output')

    return parser.parse_args()


def dem_processing(options, block_shape=(2048, 2048)):
    """Block process the input DEM and calculate slope."""
    # create output file path
    os.makedirs(options.path_out_base, exist_ok=True)
    path_out = os.path.join(os.path.normpath(options.path_out_base), 'slope_wood-provision.tif')

    # Exit if file already exists and overwrite flag is not set
    if os.path.exists(path_out) and not options.overwrite:
        logger.critical('Output file already exists. Set the -z flag to overwrite.')
        sys.exit(2)

    logger.info('* start the block processing...')
    try:
        with rasterio.open(options.path_in) as src, \
            rasterio.open(path_out, 'w', **dict(src.profile,
                                                tiled=True,
                                                compress='LZW',
                                                blockysize=256,
                                                blockxsize=256,
                                                dtype=rasterio.float32,
                                                count=1,
                                                nodata=-9999)) as dst:
            # set tags
            metadata = geo.Metadata("INCA Tool", description)
            metadata.update_dataset_tags(dst, 'Slope map for wood provision', 'percentage',
                                         src.name)
            # now iterate over the blocks of src files and process
            for index, window in geo.block_window_generator(block_shape,
                                                            dst.profile['height'], dst.profile['width']):
                aOriginal = src.read(1, window=window)
                geotransform = src.transform

                rda = rd.rdarray(aOriginal, no_data=-9999, geotransform=geotransform)
                slope = rd.TerrainAttribute(rda, attrib='slope_riserun')

                dst.write(slope, 1, window=window)

        logger.debug('done.')
    except Exception as e:
        # delete output file to rerun after error
        if os.path.exists(path_out):
            os.remove(path_out)
        logger.error('Failed to generate slope map: "%s"', e)
        sys.exit(1)


if __name__ == '__main__':
    logger = logging.getLogger(inca.woodprovision.preprocessing.__name__)
    main()
