#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    _old_dmp_forest_generation.py -i <path_to_raster_file> -c <path_to_maes_raster> -o <path_to_output_folder> -y <year> [-n] [-z] [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for processing dmp with forest mask to match the classes needed for the wood provision module.
    Note: the input raster file is only reclassified and in no way altered in resolution, extent or EPSG.

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.woodprovision.dmp_forest_generation -i <path_to_raster_file> -c <path_to_maesfile> -o <path_to_output_folder> -y <year>
    OR
    python3 _old_preprocessing_maes.py -i <path_to_raster_file> -c <path_to_maesfile> -o <path_to_output_folder> -y <year>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.

VERSION
    1.0 (2021-10-06)
"""
################## Importing modules ##################
# standard
import logging
import optparse
import os
import sys
import time

# extra official modules
import rasterio

# own modules
import inca
import inca.common.geoprocessing as geo
############## END importing modules ##################

logger = logging.getLogger(__name__)

############## MAIN ###################################
def main():
    start_time = time.time()
    #parse the arguments
    options = parse_args()

    #ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: dmp forest generation for wood provision')

    #init the metadata
    metadata = geo.Metadata("Dr. Marcel Buchhorn for INCA project",
                                "dmp forest generation for wood provision",
                                os.path.normpath(options.path_in))
    #process the input files in chunks to save memory
    try:
        #alter the block_shape to play with the memory requirements
        dmp_forest_generation(options, metadata)
    except Exception:
        logger.exception('Error during block processing of the input files.')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA wood provision dmp_forest_generation')

def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.0")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')
    parser.add_option('-i', '--in', help='Path to input raster dmp file. Needed.', dest='path_in')
    parser.add_option('-c', '--in2', help='Path to raster reclassification file. Needed.', dest='path_maes')
    parser.add_option('-o', '--out', help='Path to output folder. Needed.', dest='path_out_base')
    parser.add_option('-y', '--year', help='Year of the land cover map. Needed.', dest='year')
    parser.add_option('-n', '--nodata', help='New nodata value.', dest='new_nodata')
    parser.add_option('-z', '--overwrite', action='store_true', default=False, help='overwrite existing output files.')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_in) or (not os.path.isfile(os.path.normpath(options.path_in))):
        parser.error("the -i argument for the input raster is missing or the given path doesn't exist!!!")
    if (not options.path_maes) or (not os.path.isfile(os.path.normpath(options.path_maes))):
        parser.error("the -c argument for the raster reclassification file is missing or the given path doesn't exist!!!")
    if (not options.year):
        parser.error("the -y argument for the year of the to convert land cover map is missing or the given path doesn't exist!!!")
    if (not options.new_nodata):
        print('Warning. No new nodata value was given. The nodata value of the input raster will be used.')
        options.new_nodata = 'tbd'

    if len(args) != 0:
        parser.error('too many arguments')

    return options

def dmp_forest_generation(options, metadata):
    ''' function to block process the original LC map and transfer to a meas raster'''

    #create output roots
    if not os.path.exists(os.path.normpath(options.path_out_base)):
        os.makedirs(os.path.normpath(options.path_out_base))



    # create output file path
    path_out = os.path.join(os.path.normpath(options.path_out_base), 'dmp_forest_{}.tif'.format(options.year))


    try:
        #setup dst profile and metadata
        profile_dst = metadata.profile
        #check if we have a new nodata value
        if options.new_nodata == 'tbd':
            #we take the input nodata value
            options.new_nodata = metadata.profile['nodata']

        profile_dst.update(tiled=True,
                           compress='LZW',
                           blockysize=256,
                           blockxsize=256,
                           count=1,
                           nodata=options.new_nodata)
        # create tags
        metadata.read_raster_tags([os.path.normpath(options.path_in)])
        tags = metadata.prepare_raster_tags(
            'dmp forest map for wood provision model for year {}'.format(
                options.year))

        #set loop breaker if file already exists and overwrite flag is not set
        if (os.path.exists(path_out)) and (options.overwrite == False):
            logger.critical(
                'Output file is already existing. Please set the -z flag for overwriting. Processing will be ended.')
            sys.exit(2)

        # counter for verbose view  --> non-esential code block


        # run block processing
        with rasterio.open(os.path.normpath(options.path_in)) as src:
            aOriginal = src.read(1)
        with rasterio.open(os.path.normpath(options.path_maes)) as src:
            aMaes = src.read(1) != 5

        aOriginal[aMaes] = 0

            # create new file
        with rasterio.open(path_out, 'w', **profile_dst) as dst:
                # set tags
                dst.update_tags(**tags)
                    # write out
                dst.write(aOriginal, 1)

        logger.debug('done.')
    except Exception:
        #here we delete file to rerun after error
        if os.path.exists(path_out):
            os.remove(path_out)
        raise IOError('generation of MAES map for year {} was not successful'.format(options.year))

if __name__ == '__main__':
    # hack to get the proper logger when running the script directly with 'python _old_dmp_forest_generation.py'
    logger = logging.getLogger(inca.__name__)
    main()
