# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.

"""Create the default mask of areas where forestry activity may take place.

The default mask used in the INCA EU-wide wood provision accounts excludes

 - Natura 2000 areas.

 - Areas with steep slopes (slope above 87.5%).
"""

import argparse
import logging

import numpy as np
import rasterio

from inca.common.geoprocessing import Metadata


logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('Slope', help='map of terrain slope as a percentage.')
    parser.add_argument('Natura2000', help='Natura 2000 area mask (pixels inside Natura 2000 areas have value 1).')
    parser.add_argument('output', help='name of output file')

    return parser.parse_args()


def main():
    args = parse_args()
    log_level = logging.DEBUG if args.verbose else logging.WARNING
    logging.basicConfig(level=log_level)
    logger.debug('Running with arguments %s %s', args.Natura2000, args.Slope)

    with rasterio.open(args.Natura2000) as ds_n2k, rasterio.open(args.Slope) as ds_slope:
        if ds_n2k.res != ds_slope.res:
            raise RuntimeError('Natura 2000 mask resolution does not match slope map resolution.')
        if ds_n2k.bounds != ds_slope.bounds:
            raise RuntimeError('Natura 2000 mask extent does not match slope map extent.')

        n2k = ds_n2k.read(1)
        slope = ds_slope.read(1)
        profile = ds_slope.profile
        profile.update(
            dtype='uint8',
            compress='lzw',
            nodata=255)

    mask = np.full(shape=n2k.shape, fill_value=1, dtype=np.uint8)
    mask[(slope > 87.5) & (slope != ds_slope.nodata)] = 0
    mask[n2k == 1] = 0

    with rasterio.open(args.output, 'w', **profile) as ds:
        metadata = Metadata('INCA Tool', 'Wood provision pre-processing')
        metadata.update_dataset_tags(ds, 'Mask designating possible forestry areas.', 'dimensionless',
                                     args.Natura2000, args.Slope)
        ds.update_tags(**metadata.prepare_raster_tags(
                                                      f' {args.Natura2000} and slope mape {args.Slope}.'))
        ds.write(mask, 1)


if __name__ == '__main__':
    main()
