# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.

"""Starting from an existing for_vol_efa TSV file, write a new for_vol_efa TSV file where the NAI-FAWS table has been
replaced by data from an input CSV file, and RMOV-FNAWS are set to 0."""

import argparse
import logging

import pandas as pd

import inca.common.estat as estat

logger = logging.getLogger('inca')


def load_nai_faws(filename):
    """Read the NAI of FAWS csv file into a DataFrame and correct any negative values."""
    nai_faws = pd.read_csv(filename, sep=';', index_col=0, comment='#')
    nai_faws.columns = [int(colname) for colname in nai_faws.columns]
    # Set negative NAI values to 'NA'
    nai_faws[nai_faws < 0] = pd.NA
    # For years with negative values, replace by average of the 2 nearest surrounding positive values in that country:
    for geo in nai_faws.index[(nai_faws.isna()).any(axis=1)]:
        logger.debug('Update nai_faws for %s', geo)
        if nai_faws.loc[geo].isna().all():  # If no values are available for this region, we can't do anything.
            logger.error('No valid NAI of FAWS values for %s.', geo)
            continue
        previous_valid = None
        missing_years = []
        new_vals = nai_faws.loc[geo].copy()
        for year, value in nai_faws.loc[geo].iteritems():
            if not pd.isna(value):
                if missing_years:
                    # we see a value after finding missing values -> we are at end of a range of missing values and
                    # should update those previous missing values
                    if previous_valid is not None:
                        new_vals[missing_years] = 0.5 * (previous_valid + value)
                    else:  # if series begins with missing values, replace all by the first correct value:
                        new_vals[missing_years] = value
                    missing_years = []  # reset list of negative years
                previous_valid = value
            else:
                missing_years.append(year)
        if missing_years:  # series ends with a range of negatives
            new_vals[missing_years] = previous_valid
        nai_faws.loc[geo] = new_vals
    return nai_faws


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('nai_faws', help='CSV file containing NAI FAWS data')
    parser.add_argument('for_vol_efa', help='for_vol_efa TSV file to edit.')
    parser.add_argument('output', help='output file')

    return parser.parse_args()


def main():
    args = parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.WARNING)

    nai_faws = load_nai_faws(args.nai_faws)
    for_vol_efa = estat.read_tsv(args.for_vol_efa)

    for_vol_efa.loc[('NAI', 'FAWS', 'THS_M3')] = pd.NA
    for_vol_efa.loc[('RMOV', 'FNAWS', 'THS_M3')] = 0.
    for geo, data in nai_faws.iterrows():
        for_vol_efa.loc[('NAI', 'FAWS', 'THS_M3', geo), :] = data
    estat.write_tsv(for_vol_efa, args.output)


if __name__ == '__main__':
    main()
