#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    preprocessing_N2K.py -i <path_to_raster_file> -c <path_to_maes_raster> -o <path_to_output_folder> -y <year> [-n] [-z] [-v,--verbose] [-h,--help] [--version]

DESCRIPTION


PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.timber.preprocessing_N2K -i <path_to_raster_file> -c <path_to_csvfile> -o <path_to_output_folder> -y <year>
    OR
    python3 preprocessing_N2K.py -i <path_to_raster_file> -c <path_to_csvfile> -o <path_to_output_folder> -y <year>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.

VERSION
    1.0 (2021-10-06)
"""
################## Importing modules ##################
# standard
import logging
import optparse
import os
import time
import subprocess

# extra official modules
import pandas as pd
import numpy as np
import geopandas as gpd

# own modules
import inca
############## END importing modules ##################

logger = logging.getLogger(__name__)

############## MAIN ###################################
def main():
    start_time = time.time()
    #parse the arguments
    options = parse_args()

    #ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: preprocessing N2K for wood provision')

    #process the input files in chunks to save memory
    try:
        #alter the block_shape to play with the memory requirements
        n2k_preprocessing(options)
    except Exception:
        logger.exception('Error during block processing of the input files.')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA wood provision preprocessing_N2K')

def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.0")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')
    parser.add_option('-i', '--in', help='Path to input raster N2K shapefile. Needed.', dest='path_in')
    parser.add_option('-c', '--csv', help='Path to N2K data information file. Needed.', dest='path_csv')
    parser.add_option('-o', '--out', help='Path to output folder. Needed.', dest='path_out_base')
    parser.add_option('-y', '--year', help='Year of the land cover map. Needed.', dest='year')
    parser.add_option('-n', '--nodata', help='New nodata value.', dest='new_nodata')
    parser.add_option('-z', '--overwrite', action='store_true', default=False, help='overwrite existing output files.')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_in) or (not os.path.isfile(os.path.normpath(options.path_in))):
        parser.error("the -i argument for the input raster is missing or the given path doesn't exist!!!")
    if (not options.path_csv) or (not os.path.isfile(os.path.normpath(options.path_csv))):
        parser.error("the -c argument for the N2K information file is missing or the given path doesn't exist!!!")
    if (not options.year):
        parser.error("the -y argument for the year of the to convert land cover map is missing or the given path doesn't exist!!!")
    if (not options.new_nodata):
        print('Warning. No new nodata value was given. The nodata value of the input raster will be used.')
        options.new_nodata = 'tbd'

    if len(args) != 0:
        parser.error('too many arguments')

    return options

def n2k_preprocessing(options):
    ''' function to block process the original LC map and transfer to a meas raster'''

    #create output roots
    if not os.path.exists(os.path.normpath(options.path_out_base)):
        os.makedirs(os.path.normpath(options.path_out_base))


    try:
        # create output file path
        path_out = os.path.join(os.path.normpath(options.path_out_base), 'N2K_{}.tif'.format(options.year))

        gdf = gpd.read_file(options.path_in)
        gdf = gdf[['SITECODE', 'geometry']]

        #convert csv to something more usefull
        csv = pd.read_csv(options.path_csv, usecols=['SITECODE', 'SITETYPE', 'DATE_SPA', 'DATE_SAC', 'DATE_PROP_SCI'])
        csv.DATE_PROP_SCI = csv.DATE_PROP_SCI.fillna(csv.DATE_SAC)
        csv['date'] = np.nan
        csv.loc[csv['SITETYPE'] == 'A', 'date'] = csv.loc[csv['SITETYPE'] == 'A', 'DATE_SPA']
        csv.loc[csv['SITETYPE'].isin(['B', 'C']), 'date'] = csv.loc[csv['SITETYPE'].isin(['B', 'C']), 'DATE_PROP_SCI']
        csv = csv.filter(items=['SITECODE', 'date'])
        output = csv[csv['date'] < (((int(options.year)+1)*10000 + 101)*10**6)]

        #Filter and wright out data.
        gdf_total = gdf.merge(output, how='inner', on='SITECODE')

        rasterize(gdf_total, path_out, nodata_value=0, dtype='Byte')






        logger.debug('done.')
    except Exception:
        #here we delete file to rerun after error
        if os.path.exists(path_out):
            os.remove(path_out)
        raise IOError('generation of MAES map for year {} was not successful'.format(options.year))

def rasterize(gdb, path_out, nodata_value=0, dtype='Float32'):
    """ this function rasterize a given geopandas dataframe with a specified Column_name
         optional also give nodata value and dtype (dtype as string)"""
    # write the geodatabase to file

    temp_out = os.path.join(os.path.dirname(path_out), 'vector_{}.shp'.format(os.path.basename(path_out).split('.')[0]))

    cmd = 'gdal_rasterize -l {} -burn 1.0 -init {} -a_nodata {} -co COMPRESS=LZW -ot {} -te {} {} {} {} -tr {} {} {} {}'.format(
        os.path.basename(temp_out).split('.shp')[0],
        nodata_value,
        nodata_value,
        dtype,
        "2549000.0",
        "1402000.0",
        "6558000.0",
        "5447000.0",
        1000.0,
        1000.0,
        temp_out,
        path_out)

    if not os.path.exists(path_out):
        try:

            gdb.to_file(temp_out)
            subprocess.check_call(cmd, shell=True)
        except Exception:
            raise OSError('Could not rasterize needed vector file')
    else:
        pass


if __name__ == '__main__':
    # hack to get the proper logger when running the script directly with 'python preprocessing_N2K.py'
    logger = logging.getLogger(inca.__name__)
    main()
