#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    preprocessing_dmp_proxy.py -i <folder_path> -o <folder_path> -y <year> -r <reference_file> [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing of the DMP data out of the PROBAV or SENTINEL-3 archive.
    The total DMP of the year is calculated out of all input data in a folder. Then the raster is
    resampled plus warped to the needed AOI and resolution.
    Important: the 10daily input DMP data must be geotiffs (see lower comment for conversion).

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.woodprovision.preprocessing_dmp -i <folder_path> -o <folder_path> -y <year> -r <ref_file>
    OR
    python3 preprocessing_dmp_proxy.py -i <folder_path> -o <folder_path> -y <year> -r <ref_file>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Bert De Roo <bert.deroo@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.

VERSION
    1.0 (2021-10-06)
"""


'''
If you do not have 10-daily DMP data then you have to extract it from the CGLS:
- I use the cgls-geotiff package from VITO to access the right data and transform from netcdf to geotiff
1. clone the repository --> https://git.vito.be/scm/biggeo/cgls-geotiff.git
2. import it in folder on the MEP (e.g. ~/workspace/cgls_geotiff )
3. go to the root folder and do an editable install: python3 -m pip install --user -e .
4. check out the branch "dmp-inca" 
5. now you can call in the terminal the processing of the nc files
e.g. python3 DMP_preparation_INCA.py BioPar_DMP_V2_Global 20000101 20001231 /data/inca_vol1/input_data/CGL_BioPar_DMP_V2_Global True
'''
################## Importing modules ##################
# standard
import logging
import optparse
import os
import time
import subprocess
from datetime import datetime, timedelta

# extra official modules
import rasterio
import numpy as np

# own modules
import inca
import inca.common.geoprocessing as geo

############## END importing modules ##################

logger = logging.getLogger(__name__)


############## MAIN ###################################
def main():
    start_time = time.time()
    # parse the arguments
    options = parse_args()

    # ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA wood provision preprocessing')

    # init the metadata and raster extent of project
    metadata = geo.Metadata("Dr. Bert De Roo for INCA project",
                            "wood provision pre-processing",
                            os.path.normpath(options.path_ref))
    os.makedirs(os.path.normpath(options.path_out), exist_ok=True)

    # run the total dmp generation
    logger.info('generating the total dmp in original resolution')
    try:
        path_dmp, scale_factor = processing_total_dmp(options, metadata)
    except Exception:
        logger.exception('Error during processing of the datasets to generate the total dmp')
        raise

    # run the warp to AOI
    logger.info('bring the final DMP product to AOI')
    try:
        bring_2_AOI(options, metadata, path_dmp, scale_factor)
    except Exception:
        logger.exception('Error during processing to bring dataset to final AOI and resolution')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA wood provision DMP preprocessing')

def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.0")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')

    parser.add_option('-r', '--path_reference', help='Path to raster reference file for resampling. Needed.', dest='path_ref')
    parser.add_option('-i', '--path_in', help='Path to folder with DMP input data. Needed.', dest='path_in')
    parser.add_option('-o', '--path_out', help='Path to output folder. Needed.', dest='path_out')
    parser.add_option('-y', '--year', help='Processing year. Needed.', dest='pYear')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_ref) or (not os.path.isfile(os.path.normpath(options.path_ref))):
        parser.error("the -r argument for the reference raster is missing or the given path doesn't exist!!!")
    if (not options.path_in) or (not os.path.isdir(os.path.normpath(options.path_in))):
        parser.error("the -i argument for DMP input data is missing or the given path doesn't exist!!!")
    if (not options.pYear) or (not options.pYear.isdigit()):
        parser.error("the -y argument must be a year.")

    if len(args) != 0:
        parser.error('too many arguments')

    return options

def bring_2_AOI(options, metadata, path_dmp, scale_factor):
    ''' function of warp DMP original dataset to correct EPSG, resolution and AOI

    Note: I'm not using the function from the geoprocessing.py due to needed adaptations
    ToDo: improve the geoprocesing.py functions for warping to run also stand-alone'''

    #final path_out
    path_out = os.path.join(os.path.normpath(options.path_out), 'DMP_{}_annual_total_INCA_{}m_EPSG{}.tif'.format(
        options.pYear,
        int(metadata.profile['transform'].a),
        metadata.profile['crs'].to_epsg()))

    cmd = 'gdalwarp --config GDAL_CACHEMAX 256 -t_srs {} -te {} {} {} {} -tr {} {} -r {} -et 0 -nomd -wo SAMPLE_STEPS=50 -wo SOURCE_EXTRA=5 -wo SAMPLE-GRID=YES -co COMPRESS=LZW -co TILED=YES -co INTERLEAVE=BAND -co BIGTIFF=YES -multi -overwrite {} {}'.format(
        metadata.profile['crs'].to_string(),
        metadata.pextent.left,
        metadata.pextent.bottom,
        metadata.pextent.right,
        metadata.pextent.top,
        metadata.profile['transform'].a,
        abs(metadata.profile['transform'].e),
        'average',
        path_dmp,
        path_out)

    try:
        subprocess.check_call(cmd, shell=True)
    except Exception:
        raise OSError('Could not warp the raster file to the AOI')

    #infuse extra metadata
    metadata.read_raster_tags([path_dmp])
    tags = metadata.prepare_raster_tags(
        'Annual total of DMP derived from CGLS datasets for the year {}. value_range = [0, 250]. scaling_factor: {}. Warped to EPSG, resolution and extent of INCA AOI'.format(
            options.pYear, scale_factor), 'fraction')

    with rasterio.open(path_out, 'r+') as dst:
        dst.update_tags(**tags)
        dst.update_tags(1, scale_factor=scale_factor)

def processing_total_dmp(options, metadata, scale_factor = 0.01, valid_range = [0, 32767]):
    '''function to calculate the total dmp out of given files in folder
       NOTE: there is done no check - all tif files in folder are used to calculate total
       standard for CGLS 1km product
       valid_range: 0 - 32767
       nodata:  -1,-2
       scaling: 0.001
    '''

    #we do not care about memory so we run all in one chunk
    #get list of all files in input folder
    lFiles = []
    for entry in os.scandir(os.path.normpath(options.path_in)):
        if entry.is_file() and entry.name.endswith(".tiff"):
            lFiles.append(entry.path)

    #get main profile data from first file
    with rasterio.open(lFiles[0]) as src:
        src_profile = src.profile
        try:
            s = float(src.tags(1)['scale_factor'])
            if (s >= 0) and (s <= 1.):
                scale_factor = s
        except:
            logger.info('use standard scale factor since no matching metadata.')
        try:
            v = [int(x) for x in (src.tags(1)['valid_range']).replace(']',' ').split(' ') if x.isnumeric()]
            if (len(v) == 2) and (v[0] < v[1]):
                valid_range = v
        except:
            logger.info('use standard valid data range since no matching metadata.')

    #ini the arrays for sum and count
    aSum = np.zeros((src_profile['height'], src_profile['width']), dtype=np.uint)
    aDays = np.zeros_like(aSum, dtype=np.uint)
    maxDays = 0


    #no we loop over all files
    for element in lFiles:
        #get data
        with rasterio.open(element) as src:
            aData = src.read(1).astype(np.uint)

            time_coverage_start = datetime.strptime(src.tags()['time_coverage_start'],  '%Y-%m-%dT%H:%M:%SZ')
            time_coverage_end = datetime.strptime(src.tags()['time_coverage_end'],  '%Y-%m-%dT%H:%M:%SZ')
        #get valid data
        mValid = (aData >= valid_range[0]) & (aData <= valid_range[1])
        #fill prepared arrays

        aSum[mValid] += aData[mValid]*(time_coverage_end-time_coverage_start + timedelta(seconds=1)).days
        aDays[mValid] += (time_coverage_end-time_coverage_start + timedelta(seconds=1)).days
        maxDays += (time_coverage_end-time_coverage_start + timedelta(seconds=1)).days

    #correcting for missing data with yearly average
    #aSum[aDays < maxDays] = aSum[aDays < maxDays]/aDays[aDays < maxDays]*maxDays



    #setup dst profile
    src_profile.update(count=1, dtype=rasterio.int32, nodata=-1)

    #output filename
    path_out = os.path.join(os.path.normpath(options.path_out), 'DMP_{}_annual-total_INCA.tif'.format(options.pYear))

    #setup tags
    metadata.read_raster_tags(lFiles)
    tags = metadata.prepare_raster_tags(
        'Annual total of DMP derived from CGLS datasets for the year {}. value_range = [0, 11959955]. scaling_factor: {}'.format(
            options.pYear, scale_factor), 'dmp')

    try:
        with rasterio.open(path_out, 'w', **src_profile) as dst:
            dst.update_tags(**tags)
            dst.update_tags(1, scale_factor=scale_factor)
            dst.write(aSum.astype(src_profile['dtype']), 1)
    except:
        if os.path.exists(path_out):
            os.remove(path_out)
        raise IOError('generation of total annual DMP for year {} was not successful'.format(options.pYear))
    return path_out, scale_factor

if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger('inca.woodprovision.preprocessing_dmp')
    main()
